import listen from 'test-listen';
import fetch from 'isomorphic-unfetch';
import _$http from 'http';
const {
  createServer
} = _$http;
import _$url from 'url';
const {
  parse: parseUrl
} = _$url;

/**
 * Uses Next's internal `apiResolver` to execute api route handlers in a
 * Next-like testing environment.
 *
 * `requestPatcher/responsePatcher` are functions that receive an
 * IncomingMessage and ServerResponse object respectively. Use these functions
 * to edit the request and response before they're injected into the api
 * handler.
 *
 * `params` are passed directly to the api handler and represent processed
 * dynamic routes. This should not be confused with query string parsing, which
 * is handled automatically.
 *
 * `handler` is the actual api handler under test. It should be an async
 * function that accepts NextApiRequest and NextApiResult objects (in that
 * order) as its two parameters.
 *
 * `test` should be a function that returns a promise (or async) where test
 * assertions can be run. This function receives one parameter: fetch, which is
 * unfetch's `fetch(...)` function but with the first parameter omitted.
 */
export async function testApiHandler({
  requestPatcher,
  responsePatcher,
  paramsPatcher,
  params,
  url,
  handler,
  test
}) {
  let server = null;

  try {
    /* eslint-disable import/no-unresolved, @typescript-eslint/prefer-ts-expect-error, @typescript-eslint/ban-ts-comment */
    // ? The following is for next@>=11.1.0:
    // @ts-ignore: conditional import for earlier next versions
    const {
      apiResolver
    } = await import('next/dist/server/api-utils.js') // ? The following is for next@<11.1.0 >=10:
    // @ts-ignore: conditional import for earlier next versions
    .catch(() => import('next/dist/next-server/server/api-utils.js')) // ? The following is for next@<10:
    // @ts-ignore: conditional import for earlier next versions
    .catch(() => import('next-server/dist/server/api-utils.js'))
    /* eslint-enable import/no-unresolved */
    .catch(() => {
      throw new Error('dependency resolution failed: NTARH and the installed version of Next.js are incompatible');
    });
    const localUrl = await listen(server = createServer((req, res) => {
      url && (req.url = url);
      requestPatcher && requestPatcher(req);
      responsePatcher && responsePatcher(res);
      const finalParams = { ...parseUrl(req.url || '', true).query,
        ...params
      };
      paramsPatcher && paramsPatcher(finalParams);
      /**
       *? From next internals:
       ** apiResolver(
       **    req: IncomingMessage,
       **    res: ServerResponse,
       **    query: any,
       **    resolverModule: any,
       **    apiContext: __ApiPreviewProps,
       **    propagateError: boolean
       ** )
       */

      void apiResolver(req, res, finalParams, handler, // eslint-disable-next-line @typescript-eslint/no-explicit-any
      undefined, true);
    }));
    await test({
      fetch: init => fetch(localUrl, init)
    });
  } finally {
    var _server;

    (_server = server) === null || _server === void 0 ? void 0 : _server.close();
  }
}